from ..tools import MISSING, _or, _default, _none
from ..errors import InvalidLength, WrongType
from .errors import CallbackMissingContextCommandParameters, MissingOptionParameter, NoAsyncCallback, OptionalOptionParameter

import discord
from discord.errors import InvalidArgument

import typing
import inspect
from enum import IntEnum

def format_name(value):
    return str(value).lower().replace(" ", "-")

class SlashOption():
    """An option for a slash command
        
        Parameters
        ----------
        argument_type: :class:`int` | :class:`str` | :class:`class`
            What type of parameter the option should accept
        name: :class:`str`
            1-32 lowercase character name for the option
        description: :class:`str`, optional
            1-100 character description of the command; default name
        required: :class:`bool`, optional
            If the parameter is required or optional; default False
        choices: List[:class:`dict`], optional
            Choices for string and int types for the user to pick from; default None
                Choices should be formated like this: ``[{"name": "name of the choice", "value": "the real value"}, ...]``

                .. note::

                    The choice value has to be of the same type as the type this option accepts

        options: List[:class:`~SlashOption`]
            This parameter is only for subcommands to work, you shouldn't need to use that, unless you know what you're doing 
        """
    def __init__(self, argument_type, name, description=None, required=False, choices=None, options=None) -> None:
        """
        Creates a new option for a slash command

        Example
        ```py
        SlashOption(argument_type=int, name="Your number", required=True, choices=[{"name": "a cool number", "value": 2}])
        ```
        """
        self._json = {}
        self.argument_type = argument_type
        self.name = name
        self.description = _or(description, self.name)
        self.required = required
        self.options = _default([], options)
        self.choices = _default([], choices)
    def __repr__(self) -> str:
        return f"<discord_ui.SlashOption({str(self.to_dict())})>"
    def __eq__(self, o: object) -> bool:
        if isinstance(o, SlashOption):
            return (self.argument_type == o.argument_type and self.name == o.name and self.description == o.description and self.required == o.required and self.choices == o.choices and self.options == o.options)
        elif isinstance(o, dict):
            return (self.argument_type == o["type"] and self.name == o["name"] and self.description == o.get("description") and self.required == o.get("required", False) and self.choices == o.get("choices", []) and self.options == o.get("options", []))
        return False
    def __ne__(self, o: object) -> bool:
        return not self.__eq__(o)

    @property
    def argument_type(self) -> int:
        """Parameter type that the option accepts
        
        :type: :class:`int`
        """
        return self._json["type"]
    @argument_type.setter
    def argument_type(self, value):
        self._json["type"] = OptionType.any_to_type(value)

    @property
    def name(self) -> str:
        """The name of the option appearing in discord
        
        :type: :class:`str`
        """
        return self._json["name"]
    @name.setter
    def name(self, value):
        if len(value) > 32 or len(value) < 1:
            raise InvalidLength("name", 1, 32)
        self._json["name"] = value.lower().replace(" ", "_")

    @property
    def description(self) -> str:
        """The description of the option appearing under the name
        
        :type: :class:`str`
        """
        return self._json["description"]
    @description.setter
    def description(self, value):
        if len(value) > 100 or len(value) < 1:
            raise InvalidLength("description", 1, 100)
        self._json["description"] = value

    @property
    def required(self) -> bool:
        """Whether this parameter is required to use or not

        :type: :class:`bool`
        """
        return self._json.get("required", False)
    @required.setter
    def required(self, value):
        if not isinstance(value, bool):
            raise WrongType("required", value, "bool") 
        self._json["required"] = value

    @property
    def choices(self) -> typing.List[dict]:
        """Choices for string and int types for the user to pick from
        
        .. note::
    
            Choices are formated like this: ``[{"name": "name of the choice", "value": "the real value"}, ...]``
    
        :type: List[:class:`dict`]
        """
        return self._json.get("choices")
    @choices.setter
    def choices(self, value):
        if isinstance(value, list):
            if all(isinstance(x, dict) for x in value):
                self._json["choices"] = value
            elif all(isinstance(x, tuple) for x in value):
                self._json["choices"] = [{"name": x[0], "value": x[1]} for x in value]
            else:
                raise WrongType("choices", value, ["List[dict]", "List[tuple]"])

    @property
    def options(self) -> typing.List['SlashOption']:
        """The parameters for the command

        :type: List[:class:`~SlashOptions`], optional
        """
        return [SlashOption._from_data(x) for x in self._json.get("options", [])]
    @options.setter
    def options(self, options):
        if not isinstance(options, list):
            raise WrongType("options", options, "list")
        if all(isinstance(x, (SlashOption, dict)) for x in options):
            self._json["options"] = [(x.to_dict() if type(x) is SlashOption else x) for x in options]
        else:
            i = 0
            for x in options:
                if not isinstance(x, SlashOption) and not isinstance(x, dict):
                    raise WrongType("options[" + str(i) + "]", x, ["dict", "SlashOption"])
                i += 1

    @staticmethod
    def _from_data(data: dict):
        return SlashOption(data["type"], data["name"], data["description"], data.get("required", False), data.get("choices"), data.get("options"))

    def to_dict(self):
        return self._json

class OptionType:
    """The list of possible slash command option types"""

    SUB_COMMAND             =          Subcommand           =           1
    SUB_COMMAND_GROUP       =          Subcommand_group     =           2
    STRING                  =          String               =           3
    INTEGER                 =          Integer              =           4
    BOOLEAN                 =          Boolean              =           5
    MEMBER     =   USER     =          Member               =  User =   6
    CHANNEL                 =          Channel              =           7
    ROLE                    =          Role                 =           8
    MENTIONABLE             =          Mentionable          =           9
    FLOAT                   =          Float                =          10

    @classmethod
    def any_to_type(cls, whatever):
        """Converts something to a option type if possible"""
        if isinstance(whatever, int) and whatever in range(1, 11):
            return whatever
        if inspect.isclass(whatever):
            if whatever is str:
                return cls.STRING
            if whatever is int:
                return cls.INTEGER
            if whatever is bool:
                return cls.BOOLEAN
            if whatever in [discord.User, discord.Member]:
                return cls.MEMBER
            if whatever is discord.TextChannel:
                return cls.CHANNEL
            if whatever is discord.Role:
                return cls.ROLE
            if whatever is float:
                return cls.FLOAT
        if isinstance(whatever, str):
            whatever = whatever.lower()
            if whatever in ["str", "string"]:
                return cls.STRING
            if whatever in ["int", "integer"]:
                return cls.INTEGER
            if whatever in ["bool", "boolean"]:
                return cls.BOOLEAN
            if whatever in ["user", "discord.user", "member", "discord.member", "usr", "mbr"]:
                return cls.MEMBER
            if whatever in ["channel", "textchannel", "discord.textchannel", "txtchannel"]:
                return cls.CHANNEL
            if whatever in ["role", "discord.role"]:
                return cls.ROLE
            if whatever in ["mentionable", "mention"]:
                return cls.MENTIONABLE
            if whatever in ["float", "floating", "floating number", "f"]:
                return cls.FLOAT
class AdditionalType:
    MESSAGE     =       44
    GUILD       =       45
class CommandType(IntEnum):
    Slash       =              1
    User        =              2
    Message     =              3

    @staticmethod
    def from_string(typ):
        if isinstance(typ, str):
            if typ.lower() == "slash":
                return CommandType.Slash
            elif typ.lower() == "user":
                return CommandType.User
            elif typ.lower() == "message":
                return CommandType.Message
        elif isinstance(typ, CommandType):
            return typ
        else:
            return CommandType(typ)
    def __str__(self):
        return self.name

class SlashPermission():
    """Permissions for a slash commannd
        
        Parameters
        ----------
            allowed: :class:`dict` | List[:class:`discord.Member` | :class:`discord.User` | :class:`discord.Role`], optional
                A list of ids, users or members that can use the command; default None
                    Format: ``{"role_or_user_id": permission_type}]``


            forbidden: :class:`dict` | List[:class:`discord.Member` | :class:`discord.User` | :class:`discord.Role`], optional
                A list of ids, users or members that are forbidden to use the command; default None
        

            .. note::

                If you want to use a role id, the permission type has to be 1, and if you want to specify a user id, it has to be 2

                You can use ``SlashPermission.ROLE`` and ``SlashPermission.USER`` instead

        """
    def __init__(self, allowed: dict=None, forbidden=None) -> None:
        """Creates a new permission object for a slash command
        
        Example
        ```py
        SlashPermission(allowed=[
                await bot.fetch_user(bot.owner_id)
            ], forbidden={
                539459006847255232: SlashPermission.User,
                874357255829606402: SlashPermission.Role
            }
        )
        ```
        """
        
        self._json = []
        if allowed is not None:
            if isinstance(allowed, dict):
                for _id, _type in allowed.items():
                    self._json.append(
                        {
                            "id": int(_id),
                            "type": _type,
                            "permission": True
                        }
                    )
            elif isinstance(allowed, list):
                for t in allowed:
                    self._json.append({
                        "id": t.id,
                        "type": SlashPermission.USER if isinstance(t, (discord.User, discord.Member)) else SlashPermission.ROLE,
                        "permission": True
                    })
        if forbidden is not None:
            if isinstance(forbidden, dict):
                for _id, _type in forbidden.items():
                    self._json.append(
                        {
                            "id": int(_id),
                            "type": _type,
                            "permission": False
                        }
                    )
            elif isinstance(forbidden, list):
                for t in forbidden:
                    self._json.append({
                        "id": t.id,
                        "type": SlashPermission.USER if type(t) in [discord.User, discord.Member] else SlashPermission.ROLE,
                        "permission": False
                    })
            else:
                raise WrongType("forbidden", forbidden, ["Dict[int | str, int]", "List[discord.Member | discord.User | discord.Role]"])

    @classmethod
    def Empty(cls) -> 'SlashPermission':
        """Returns an empty permission for the command"""
        return cls()
    def to_dict(self):
        return self._json
    def __eq__(self, o: object) -> bool:
        if isinstance(o, SlashPermission):
            return (
                len(self.allowed) == len(o.allowed) and all(self.allowed[i] == o.allowed[i] for i, _ in enumerate(o.allowed)) and
                len(self.forbidden) == len(o.forbidden) and all(self.forbidden[i] == o.forbidden for i, _ in enumerate(o.forbidden))
            )
        elif isinstance(o, dict):
            o_allowed = [x for x in o["permissions"] if x["permission"] == True]
            o_forbidden = [x for x in o["permissions"] if x["permission"] == False]

            return (
                len(self.allowed) == len(o_allowed) and all(self.allowed[i] == o_allowed[i] for i, _ in enumerate(self.allowed)) and
                len(self.forbidden) == len(o_forbidden) and all(self.forbidden[i] == o_forbidden[i] for i, _ in enumerate(self.forbidden))
            )
        return False
    def __ne__(self, o: object) -> bool:
        return not self.__eq__(o)
    def __repr__(self) -> str:
        return f"<discord_ui.SlashPermission({self.to_dict()})>"

    ROLE        =       Role      =   1
    USER        =       User      =   2

    @property
    def allowed(self) -> dict:
        return [x for x in self._json if x["permission"] == True]
    @property
    def forbidden(self) -> dict:
        return [x for x in self._json if x["permission"] == False]


class BaseCommand():
    __slots__ = ('__aliases__', '__sync__', '__original_name__', '__auto_defer__')
    def __init__(self, command_type, callback, name=None, description=None, options=None, guild_ids=None, default_permission=None, guild_permissions=None) -> None:
        self.__aliases__ = getattr(callback, "__aliases__", None)
        self.__sync__ = getattr(callback, "__sync__", True)
        self.__auto_defer__ = getattr(callback, "__auto_defer__", None)
        self.__guild_changes__ = getattr(callback, "__guild_changes__", {})

        self._json = {"type": getattr(command_type, "value", command_type)}

        self.options = _default([], options)
        if callback is not None:
            if not inspect.iscoroutinefunction(callback):
                raise NoAsyncCallback()

            callback_params = inspect.signature(callback).parameters
            if self.options is not None:
                for op in self.options:
                    if callback_params.get(op.name) is None:
                        raise MissingOptionParameter(op.name)
                    param = callback_params[op.name]
                    if not op.required and param.default is param.empty:
                        raise OptionalOptionParameter(param.name)
            if self.options in [[], None] and self.command_type is CommandType.Slash:
                _ops = []
                has_self = False
                for _i, _name in enumerate(callback_params):
                    # ignore context parameter
                    if _name == "self":
                        has_self = True
                        continue
                    if _i == (1 if has_self else 0):
                        continue
                    _val = callback_params.get(_name)
                    op_type = None
                    if _val.annotation != _val.empty:
                        op_type = _val.annotation
                    elif _val.default != inspect._empty:
                        op_type = type(_val.default)
                    else: 
                        op_type = _name
                    if OptionType.any_to_type(op_type) is None:
                        raise InvalidArgument("Could not find a matching option type for parameter '" + str(op_type) + "'")
                    _ops.append(SlashOption(op_type, _name, required=_val.default == inspect._empty))
                self.options = _ops

        self.callback: function = callback
        self.name = _or(name, self.callback.__name__ if not _none(self.callback) else None)
        # Set the original name to the name once so if the name should be changed, this value still stays to what it is
        self.__original_name__ = self.name
        self.description = _or(description, inspect.getdoc(callback).split("\n")[0] if not _none(callback) and inspect.getdoc(callback) is not None else None, self.name)
        if default_permission is None:
            default_permission = True
        self.default_permission: bool = default_permission
        if not _none(guild_permissions):
            for _id, perm in list(guild_permissions.items()):
                if not isinstance(_id, (str, int, discord.User, discord.Member, discord.Role)):
                    raise WrongType("guild_permissions key " + str(_id), _id, ["str", "int", "discord.User", "discord.Member", "discord.Role"])
                if not isinstance(perm, SlashPermission):
                    raise WrongType("guild_permission[" + ("'" if isinstance(_id, str) else "") + str(_id) + ("'" if isinstance(_id, str) else "") + "]", perm, "SlashPermission")
        
        self.guild_permissions: typing.Dict[(typing.Union[str, int], SlashPermission)] = guild_permissions
        self.permissions: SlashPermission = SlashPermission()
        self.guild_ids: typing.List[int] = _default(None, [int(x) for x in _or(guild_ids, [])])
        """The ids of the guilds where the command is available"""
    def __str__(self) -> str:
        return str(self.to_dict())
    def __eq__(self, o: object) -> bool:
        if isinstance(o, dict):
            return (
                o.get('type') == self.command_type.value 
                and o.get('name') == self.name
                and o.get('description') == self.description
                and o.get('options', []) == self.options
                and o.get("default_permission", True) == self.default_permission
            )
        elif isinstance(o, SlashCommand):
            return (
                o._json('type') == self._json["type"]
                and o.name == self.name
                and o.description == self.description
                and o.options == self.options
                and o.default_permission == self.default_permission
            )
        else:
            return False
    def __ne__(self, o: object) -> bool:
        return not self.__eq__(o)

    @property
    def is_message_context(self) -> bool:
        """Wether is command is a message-context command"""
        return self.command_type is CommandType.Message
    @property
    def is_user_context(self) -> bool:
        """Whether this command is a user-context command"""
        return self.command_type is CommandType.User
    @property
    def is_chat_input(self) -> bool:
        """Whether this command is a chat-input command"""
        return self.command_type is CommandType.Slash
    @property
    def is_slash_command(self) -> bool:
        """
        Whether this command is a slashcommand
            Same as `.is_chat_input`
        """
        return self.is_chat_input
    
    @property
    def original_name(self) -> str:
        """The original name for this command"""
        return self.__original_name__
    @property
    def aliases(self) -> typing.List[str]:
        """The list of available aliases for this command"""
        return self.__aliases__
    @property
    def has_aliases(self) -> bool:
        """Returns True if this command has aliases"""
        return hasattr(self, "__aliases__") and self.__aliases__ is not None
    @property
    def is_alias(self) -> bool:
        """Returns True if this command is an alias to another command"""
        return self.__aliases__ is not None and self.name in self.__aliases__

    # region command
    @property
    def command_type(self) -> CommandType:
        return CommandType(self._json["type"])

    @property
    def name(self) -> str:
        """The name of the slash command
        
        :type: :class:`str`
        """
        return self._json["name"]
    @name.setter
    def name(self, value):
        if _none(value):
            raise InvalidArgument("You have to specify a name")
        if not isinstance(value, str):
            raise WrongType("name", value, "str")
        if len(value) > 32 or len(value) < 1:
            raise InvalidLength("name", 1, 32)
        if self.command_type is CommandType.Slash:
            self._json["name"] = format_name(value)
        else:
            self._json["name"] = value
    @property
    def description(self) -> str:
        """The description of the command
        
        :type: :class:`str`
        """
        return self._json['description']
    @description.setter
    def description(self, value):
        if not isinstance(value, str):
            raise WrongType("description", value, "str")
        if len(value) > 100 or len(value) < 1:
            raise InvalidLength("description", 1, 100)
        self._json["description"] = value
    @property
    def options(self) -> typing.List['SlashOption']:
        """The parameters for the command
        
        :type: List[:class:`~SlashOption`]
        """
        return [SlashOption._from_data(x) for x in self._json.get("options", [])]
    @options.setter
    def options(self, options):
        if not isinstance(options, list):
            raise WrongType("options", options, "list")
        if all(isinstance(x, (SlashOption, dict)) for x in options):
            self._json["options"] = [(x.to_dict() if isinstance(x, SlashOption) else x) for x in options]
        else:
            i = 0
            for x in options:
                if not isinstance(x, (SlashOption, dict)):
                    raise WrongType("options[" + str(i) + "]", x, ["dict", "SlashOption"])
                i += 1
    # endregion
    # region permissions
    @property
    def default_permission(self) -> bool:
        """Whether this command can be used by default or not
        
        :type: :class:`bool`
        """
        return self._json.get("default_permission", False)
    @default_permission.setter
    def default_permission(self, value):
        self._json["default_permission"] = value
    # endregion

    def _patch(self, command):
        self.__aliases__ = command.__aliases__
        self.guild_permissions = command.guild_permissions

    def copy(self):
        """Copies itself into a new object"""
        raise NotImplementedError()

    def to_dict(self):
        return self._json

class SlashCommand(BaseCommand):
    """A basic slash command
        
        Parameters
        ----------
            callback: :class:`function`
                The callback function
            name: :class:`str`
                1-32 characters long name
                    If name is not passed, the name of the callback function will be used; default None
                    
                .. important::

                    The name will be corrected automaticaly (spaces will be replaced with "-" and the name will be lowercased)

            description: :class:`str`, optional
                1-100 character description of the command; default None
                    If description is not passed, the docstring description of the callback function is used. 
                    If no docstring exists, the name of the command will be used
            options: List[:class:`~SlashOptions`], optional
                Parameters for the command; default None
            choices: List[:class:`tuple`] | List[:class:`dict`], optional
                Choices for string and int types for the user to pick from; default None
            guild_ids: :class:`str` | :class:`int`, optional
                A list of guild ids where the command is available; default None
            default_permission: :class:`bool`
                Whether the command should be usable for everyone or not
            guild_permissions: Dict[``guild_id``: :class:`~SlashPermission`]
                The permissions for the command in guilds
                    Format: ``{"guild_id": SlashPermission}``


    """
    def __init__(self, callback, name=None, description=None, options=None, guild_ids=None, default_permission=None, guild_permissions=None) -> None:
        """
        Creates a new base slash command
        
        Example

        ```py
        async def my_function(command, parameter=None):
            pass

        SlashCommand(callback=my_function, name="hello_world", description="This is a test command",
            options=[
                SlashOption(str, name="parameter", description="this is a parameter", choices=[("choice 1", 1)])
            ], guild_ids=[785567635802816595], default_permission=False,
            guild_permissions={
                785567635802816595: SlashPermission(allowed={"539459006847254542": SlashPermission.USER})
            })
        ```
        """
        BaseCommand.__init__(self, CommandType.Slash, callback, name, description, options, guild_ids, default_permission, guild_permissions)
    def copy(self) -> 'SlashCommand':
        c = SlashCommand(self.callback, self.name, self.description, self.options, self.guild_ids, self.default_permission, self.guild_permissions)
        for x in self.__slots__:
            setattr(c, x, getattr(self, x))
        return c

class SlashSubcommand(BaseCommand):
    def __init__(self, callback, base_names, name, description=None, options=[], guild_ids=None, default_permission=None, guild_permissions=None) -> None:
        if isinstance(base_names, str):
            base_names = [base_names]
        if len(base_names) > 2:
            raise InvalidArgument("subcommand groups are currently limited to 2 bases")
        if any([len(x) > 32 or len(x) < 1 for x in base_names]):
            raise InvalidLength("base_names", 1, 32)
        BaseCommand.__init__(self, CommandType.Slash, callback, name, description, options, guild_ids=guild_ids, default_permission=default_permission, guild_permissions=guild_permissions)
        self.base_names = [format_name(x) for x in base_names]
    
    def to_option(self) -> SlashOption:
        return SlashOption(OptionType.SUB_COMMAND, self.name, self.description, options=self.options or None, required=False)
    def to_dict(self):
        return self.to_option().to_dict()
    def copy(self):
        c = SlashSubcommand(self.callback, self.base_names, self.name, self.description, self.options, self.guild_ids, self.default_permission, self.guild_permissions)
        for x in self.__slots__:
            setattr(c, x, getattr(self, x))
        return c

class ContextCommand(BaseCommand):
    def __init__(self, context_type, callback, name=None, guild_ids=None, default_permission = True, guild_permissions = None) -> None:
        if callback is not None:
            callback_params = inspect.signature(callback).parameters
            if len(callback_params) < 2:
                raise CallbackMissingContextCommandParameters()
        BaseCommand.__init__(self, context_type, callback, name=name, guild_ids=guild_ids, default_permission=default_permission, guild_permissions=guild_permissions)

    @property
    def description(self) -> str:
        return ""
    @description.setter
    def description(self, value):
        pass
    @property
    def options(self) -> list:
        return []
    @options.setter
    def options(self, value):
        pass

class UserCommand(ContextCommand):
    def __init__(self, callback, name=None, guild_ids = None, default_permission = True, guild_permissions = None) -> None:
        ContextCommand.__init__(self, CommandType.User, callback, name, guild_ids, default_permission, guild_permissions)
    def copy(self) -> 'UserCommand':
        c = UserCommand(self.callback, self.name, self.guild_ids, self.default_permission, self.guild_permissions)
        for x in self.__slots__:
            setattr(c, x, getattr(self, x))
        return c

class MessageCommand(ContextCommand):
    def __init__(self, callback, name=None, guild_ids = None, default_permission = True, guild_permissions = None) -> None:
        ContextCommand.__init__(self, CommandType.Message, callback, name, guild_ids, default_permission, guild_permissions)
    def copy(self) -> 'MessageCommand':
        c = MessageCommand(self.callback, self.name, self.guild_ids, self.default_permission, self.guild_permissions)
        for x in self.__slots__:
            setattr(c, x, getattr(self, x))
        return c


from .components import ComponentType
from .slash.types import BaseCommand, MessageCommand, SlashCommand, SlashSubcommand, UserCommand

import discord
from discord.errors import InvalidArgument
from discord.ext.commands import errors
from discord.ext.commands.cooldowns import BucketType, CooldownMapping

import asyncio
import datetime
from typing import Optional
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

class WrongListener(errors.CheckFailure):
    """
    Exception raised when a listening component received a component event that doesn't meet the check conditions

    This inherits from :exc:`CheckFailure`
    """
    def __init__(self, message: Optional[str] = None) -> None:
        super().__init__(message or "The used component doesn't this listener's check conditions.")

class BaseCallable():
    """
    A base class for all cog objects
    
    IMPORTANT: Most of this code is taken out of the discord.ext.commands module
        Ref: https://github.com/Rapptz/discord.py/blob/a2a7b0f076c763f08c871d76f9477d1d0eed1973/discord/ext/commands/core.py#L205
    """
    def __init__(self, callback) -> None:
        self.callback = callback
        self._before_invoke = None
        self._after_invoke = None
        
        # Checks for the dpy decorator
        self.__commands_checks__ = []
        if hasattr(self.callback, "__commands_checks__"):
            self.__commands_checks__ = self.callback.__commands_checks__

        cooldown = None
        if hasattr(self.callback, "__commands_cooldown__"):
            cooldown = self.callback.__commands_cooldown__
        try:
            self._buckets = CooldownMapping(cooldown)
        except TypeError:
            self._buckets = CooldownMapping(cooldown, BucketType.default)

        self._max_concurrency = None
        if hasattr(self.callback, "__commands_max_concurrency__"):
            self._max_concurrency = self.callback.__commands_max_concurrency__

        self._before_invoke = None
        try:
            before_invoke = self.callback.__before_invoke__
        except AttributeError:
            pass
        else:
            self.before_invoke(before_invoke)

        self._after_invoke = None
        try:
            after_invoke = self.callback.__after_invoke__
        except AttributeError:
            pass
        else:
            self.after_invoke(after_invoke)

        self.on_error = None
        
    async def __call__(self, *args, **kwds):
        return self.callback(*args, **kwds)
    async def invoke(self, ctx, *args, **kwargs):
        if not await self.can_run(ctx):
            raise errors.CheckFailure()
        if self._before_invoke is not None:
            await self._before_invoke(ctx)

        if self._max_concurrency is not None:
            await self._max_concurrency.acquire(ctx)
        try:
            # cooldown checks
            self._prepare_cooldowns(ctx)
        except Exception:
            if self._max_concurrency is not None:
                await self._max_concurrency.release(ctx)
            raise
        try:
            if hasattr(self, "cog"):
                await self.callback(self.cog, ctx, *args, **kwargs)
            else:
                await self.callback(ctx, *args, **kwargs)
        except Exception as ex:
            if not self.on_error:
                raise ex
            else:
                self.on_error(getattr(self, "cog", None), ctx, ex)
        if self._after_invoke is not None:
            await self._after_invoke(ctx)
    async def can_run(self, ctx):
        """Whether the command can be run"""
        predicates = self.checks
        if not predicates:
            # since we have no checks, then we just return True.
            return True
        return await discord.utils.async_all(predicate(ctx) for predicate in predicates)  # type: ignore
    @property
    def checks(self):
        return self.__commands_checks__
    def add_check(self, check):
        self.__commands_checks__.append(check)
    def remove_check(self, check):
        self.__commands_checks__.remove(check)
    def _prepare_cooldowns(self, ctx) -> None:
        if self._buckets.valid:
            dt = ctx.message.edited_at or ctx.message.created_at
            current = dt.replace(tzinfo=datetime.timezone.utc).timestamp()
            bucket = self._buckets.get_bucket(ctx.message, current)
            if bucket is not None:
                retry_after = bucket.update_rate_limit(current)
                if retry_after:
                    raise CommandOnCooldown(bucket, retry_after, self._buckets.type)  # type: ignore

    def is_on_cooldown(self, ctx) -> bool:
        """
        Checks whether the command is currently on cooldown.

        Parameters
        -----------
        ctx: :class:`.Context`
            The invocation context to use when checking the commands cooldown status.

        Returns
        --------
        :class:`bool`
            A boolean indicating if the command is on cooldown.
        """
        if not self._buckets.valid:
            return False

        bucket = self._buckets.get_bucket(ctx.message)
        dt = ctx.message.edited_at or ctx.message.created_at
        current = dt.replace(tzinfo=datetime.timezone.utc).timestamp()
        return bucket.get_tokens(current) == 0
    def reset_cooldown(self, ctx) -> None:
        """
        Resets the cooldown on this command.

        Parameters
        -----------
        ctx: :class:`.Context`
            The invocation context to reset the cooldown under.
        """
        if self._buckets.valid:
            bucket = self._buckets.get_bucket(ctx.message)
            bucket.reset()
    def get_cooldown_retry_after(self, ctx) -> float:
        """
        Retrieves the amount of seconds before this command can be tried again.

        .. versionadded:: 1.4

        Parameters
        -----------
        ctx: :class:`.Context`
            The invocation context to retrieve the cooldown from.

        Returns
        --------
        :class:`float`
            The amount of time left on this command's cooldown in seconds.
            If this is ``0.0`` then the command isn't on cooldown.
        """
        if self._buckets.valid:
            bucket = self._buckets.get_bucket(ctx.message)
            dt = ctx.message.edited_at or ctx.message.created_at
            current = dt.replace(tzinfo=datetime.timezone.utc).timestamp()
            return bucket.get_retry_after(current)

        return 0.0
    def error(self, coro):
        """
        A decorator that registers a coroutine as a local error handler.
        
        A local error handler is an :func:`.on_command_error` event limited to
        a single command. However, the :func:`.on_command_error` is still
        invoked afterwards as the catch-all.

        Parameters
        -----------
        coro: :ref:`coroutine <coroutine>`
            The coroutine to register as the local error handler.

        Raises
        -------
        TypeError
            The coroutine passed is not actually a coroutine.
        """

        if not asyncio.iscoroutinefunction(coro):
            raise TypeError('The error handler must be a coroutine.')

        self.on_error = coro
        return coro
    def has_error_handler(self) -> bool:
        """
        :class:`bool`: Checks whether the command has an error handler registered.

        .. versionadded:: 1.7
        """
        return hasattr(self, 'on_error')
    def before_invoke(self, coro):
        """
        A decorator that registers a coroutine as a pre-invoke function.

        This function is called directly before the command is
        called. This makes it a useful function to set up database
        connections or any type of set up required.

        This post-invoke function takes a sole parameter, one of :class:`~SlashedCommand` | :class:`~SlashedSubCommand` | 
        :class:`~SlashedContext` | :class:`~PressedButton` | :class:`SelectedMenu`, depending on what this is used for

        Parameters
        -----------
        coro: :ref:`coroutine <coroutine>`
            The coroutine to register as the pre-invoke hook.

        Raises
        -------
        TypeError
            The coroutine passed is not actually a coroutine.
        """
        if not asyncio.iscoroutinefunction(coro):
            raise TypeError('The pre-invoke hook must be a coroutine.')

        self._before_invoke = coro
        return coro
    def after_invoke(self, coro):
        """
        A decorator that registers a coroutine as a post-invoke function.

        A post-invoke function is called directly after the command is
        called. This makes it a useful function to clean-up database
        connections or any type of clean up required.

        This post-invoke function takes a sole parameter, one of :class:`~SlashedCommand` | :class:`~SlashedSubCommand` | 
        :class:`~SlashedContext` | :class:`~PressedButton` | :class:`SelectedMenu`, depending on what this is used for


        Parameters
        -----------
        coro: :ref:`coroutine <coroutine>`
            The coroutine to register as the post-invoke hook.

        Raises
        -------
        TypeError
            The coroutine passed is not actually a coroutine.
        """
        if not asyncio.iscoroutinefunction(coro):
            raise TypeError('The post-invoke hook must be a coroutine.')

        self._after_invoke = coro
        return coro
class BaseSlash(BaseCallable):
    def __init__(self, callback) -> None:
        self.__slots__ = BaseCommand.__slots__ + ('cog',)
        self.__type__ = 1
        BaseCallable.__init__(self, callback)

class CogCommand(BaseSlash, SlashCommand):
    def __init__(self, *args, **kwargs) -> None:
        SlashCommand.__init__(self, *args, **kwargs)
        BaseSlash.__init__(self, args[0])
    def copy(self) -> 'CogCommand':
        c = CogCommand(self.callback, self.name, self.description, self.options, self.guild_ids, self.default_permission, self.guild_permissions)
        for x in self.__slots__:
            setattr(c, x, getattr(self, x, None))
        return c
class CogSubCommandGroup(BaseSlash, SlashSubcommand):
    def __init__(self, *args, **kwargs) -> None:
        SlashSubcommand.__init__(self, *args, **kwargs)
        BaseSlash.__init__(self, args[0])
    def copy(self) -> 'CogSubCommandGroup':
        c = CogSubCommandGroup(self.callback, self.base_names, self.name, self.description, self.options, self.guild_ids, self.default_permission, self.guild_permissions)
        for x in self.__slots__:
            setattr(c, x, getattr(self, x, None))
        return c
class CogMessageCommand(BaseSlash, MessageCommand):
    def __init__(self, *args, **kwargs) -> None:
        MessageCommand.__init__(self, *args, **kwargs)
        BaseSlash.__init__(self, args[0])
    def copy(self) -> 'CogMessageCommand':
        c = CogMessageCommand(self.callback, self.name, self.guild_ids, self.default_permission, self.guild_permissions)
        for x in self.__slots__:
            setattr(c, x, getattr(self, x, None))
        return c
class CogUserCommand(BaseSlash, UserCommand):
    def __init__(self, *args, **kwargs) -> None:
        UserCommand.__init__(self, *args, **kwargs)
        BaseSlash.__init__(self, args[0])
    def copy(self) -> 'CogUserCommand':
        c = CogUserCommand(self.callback, self.name, self.guild_ids, self.default_permission, self.guild_permissions)
        for x in self.__slots__:
            setattr(c, x, getattr(self, x, None))
        return c

class ListeningComponent(BaseCallable):
    def __init__(self, callback, messages, users, component_type, check, custom_id) -> None:
        BaseCallable.__init__(self, callback)
        self.__type__ = 2
        def predicate(ctx):
            checks = []
            if messages not in [None, []]:
                checks.append(ctx.message.id in [(x.id if hasattr(x, "id") else int(x)) for x in messages])
            if users not in [None, []]:
                checks.append(ctx.author.id in [(x.id if hasattr(x, "id") else int(x)) for x in users])
            if component_type is not None:
                checks.append(ctx.component_type is (ComponentType.Button if component_type in [ComponentType.Button, "button"] else ComponentType.Select))
            if check is not None:
                checks.append(check(ctx) is True)
                if not all(checks):
                    raise WrongListener()
            return True
        self.add_check(predicate)
        self.custom_id = custom_id

def slash_cog(name=None, description=None, options=[], guild_ids=None, default_permission=None, guild_permissions=None):
    """
    A decorator for cogs that will register a slashcommand
    
    command in discord
        ``/name [options]``
    
    Parameters
    ----------
    name: :class:`str`, optional
        1-32 characters long name; default MISSING

        .. note::

            The name will be corrected automaticaly (spaces will be replaced with "-" and the name will be lowercased)
    
    description: :class:`str`, optional
        1-100 character description of the command; default the command name
    options: List[:class:`~SlashOptions`], optional
        The parameters for the command; default MISSING
    choices: List[:class:`tuple`] | List[:class:`dict`], optional
        Choices for string and int types for the user to pick from; default MISSING
    guild_ids: List[:class:`str` | :class:`int`], optional
        A list of guild ids where the command is available; default MISSING
    default_permission: :class:`bool`, optional
        Whether the command can be used by everyone or not
    guild_permissions: Dict[``guild_id``: :class:`~SlashPermission`]
        The permissions for the command in guilds
            Format: ``{"guild_id": SlashPermission}``

    Decorator
    ---------

    callback: :class:`method(ctx)` 
        The asynchron function that will be called if the command was used
        
        ctx: :class:`~SlashedCommand`
            The used slash command

        .. note::

            ``ctx`` is just an example name, you can use whatever you want for that

    Example
    --------

    .. code-block::

        class My_Cog(commands.Cog)
            ... 

            @slash_cog(name="hello_world", options=[
                SlashOption(str, name="parameter", description="this is a parameter", choices=[{ "name": "choice 1", "value": "test" }])
            ], guild_ids=[785567635802816595], default_permission=False, 
            guild_permissions={
                    785567635802816595: SlashPermission(allowed={"539459006847254542": SlashPermission.USER})
                }
            )
            async def hello_world(self, ctx, parameter = None):
                ...
    """
    def wraper(callback):
        return CogCommand(callback, name, description, options, guild_ids=guild_ids, default_permission=default_permission, guild_permissions=guild_permissions)
    return wraper
def subslash_cog(base_names, name=None, description=None, options=[], guild_ids=None, default_permission=None, guild_permissions=None):
    """
    A decorator for cogs that will register a subcommand/subcommand-group
  
    command in discord
        ``/base_names... name [options]``

    Parameters
    ----------
    base_names: List[:class:`str`] | :class:`str`
        The names of the parent bases, currently limited to 2
            If you want to make a subcommand (``/base name``), you have to use a str instead of a list
    name: :class:`str`, optional
        1-32 characters long name; default MISSING
        
        .. note::

            The name will be corrected automaticaly (spaces will be replaced with "-" and the name will be lowercased)
    description: :class:`str`, optional
        1-100 character description of the command; default the command name
    options: List[:class:`~SlashOptions`], optional
        The parameters for the command; default MISSING
    choices: List[:class:`tuple`] | List[:class:`dict`], optional
        Choices for string and int types for the user to pick from; default MISSING
    guild_ids: List[:class:`str` | :class:`int`], optional
        A list of guild ids where the command is available; default MISSING
    default_permission: :class:`bool`, optional
        Whether the command can be used by everyone or not
    guild_permissions: Dict[``guild_id``: :class:`~SlashPermission`]
        The permissions for the command in guilds
            Format: ``{"guild_id": SlashPermission}``

    .. note::

        Permissions will be the same for every subcommand with the same base

    Decorator
    ---------
    callback: :class:`method(ctx)`
        The asynchron function that will be called if the command was used
            ctx: :class:`~SlashedSubCommand`
                The used slash command

            .. note::

                ``ctx`` is just an example name, you can use whatever you want for that
    
    Example
    --------
    subcommand

    .. code-block::

        class My_Cog(commands.Cog):
            ...

            @subslash_cog(base_names="hello", name="world", options=[
                SlashOption(argument_type="user", name="user", description="the user to tell the holy words")
            ], guild_ids=[785567635802816595])
            async def command(ctx, user):
                ...

    subcommand-group

    .. code-block::

        class My_Cog(commands.Cog):
            ...

            @subslash_cog(base_names=["hello", "beautiful"], name="world", options=[
                SlashOption(argument_type="user", name="user", description="the user to tell the holy words")
            ], guild_ids=[785567635802816595])
            async def command(ctx, user):
                ...

    """
    def wraper(callback):
        return CogSubCommandGroup(callback, base_names, name, description=description, options=options, guild_ids=guild_ids, default_permission=default_permission, guild_permissions=guild_permissions)
    return wraper
def context_cog(type: Literal["user", 2, "message", 3], name=None, guild_ids=None, default_permission=None, guild_permissions=None):
    """
    Decorator for cogs that will register a context command in discord
            ``Right-click message or user`` -> ``apps`` -> ``commands is displayed here``

    Parameters
    ----------
        type: Literal[``'user'``, ``2`` | ``'message'`` | ``3``]
            The type of the contextcommand. 
                ``'user'`` and ``2`` are user-commands; ``'message'`` and ``3`` are message-commansd
        name: :class:`str`, optional
            The name of the command; default MISSING
        guild_ids: List[:class:`str` | :class:`int`]
            A list of guilds where the command can be used
        default_permission: :class:`bool`, optional
            Whether the command can be used by everyone; default True
        guild_permissions: Dict[:class:`SlashPermission`], optional
            Special permissions for guilds; default MISSING

    Decorator
    ---------

        callback: :class:`method(ctx, message)`
            The asynchron function that will be called if the command was used
                ctx: :class:`~SlashedSubCommand`
                    The used slash command
                message: :class:`~Message`
                    The message on which the command was used
                
                .. note::

                    ``ctx`` and ``message`` are just example names, you can use whatever you want for that
    
        Example
        -------
        
        .. code-block::

            class My_Cog(commands.Cog):
                ...

                # message command
                @context_cog(type="message", name="quote", guild_ids=[785567635802816595])
                async def quote(ctx, message):
                    ...
                
                # user command
                @context_cog(type="user", name="mention", guild_ids=[785567635802816595])
                async def mention(ctx, user):
                    ...
        """
    def wraper(callback):
        if type in ["user", 2]:
            return CogUserCommand(callback, name, guild_ids=guild_ids, default_permission=default_permission, guild_permissions=guild_permissions)
        elif type in ["message", 3]:
            return CogMessageCommand(callback, name, guild_ids=guild_ids, default_permission=default_permission, guild_permissions=guild_permissions)
        else:
            raise InvalidArgument("Invalid context type! type has to be one of 'user', 1, 'message', 2!")
    return wraper
def listening_component_cog(custom_id, messages=None, users=None, component_type: Literal['button', 'select']=None, check=lambda _ctx: True):
    """
    Decorator for cogs that will register a listening component

    Parameters
    ----------
        custom_id: :class:`str`
            The custom_id of the components to listen to
        messages: List[:class:`discord.Message` | :class:`int` :class:`str`], Optional
            A list of messages or message ids to filter the listening component
        users: List[:class:`discord.User` | :class:`discord.Member` | :class:`int` | :class:`str`], Optional
            A list of users or user ids to filter
        component_type: Literal[``'button'`` | ``'select'``]
            What type the used component has to be of (select: SelectMenu, button: Button)
        check: :class:`function`, Optional
            A function that has to return True in order to invoke the listening component
                The check function takes to parameters, the component and the message

    Decorator
    ---------
        callback: :class:`method(ctx)`
            The asynchron function that will be called if a component with the custom_id was invoked

            There will be one parameters passed

                ctx: :class:`~PressedButton` or :class:`~SelectedMenu`
                    The invoked component
                
                .. note::

                    ``ctx`` is just an example name, you can use whatever you want for it

    Example
    -------
    
    .. code-block::
    
        @ui.components.listening_component("custom_id", [539459006847254542], [53945900682362362])
        async def callback(ctx):
            ...
    """
    def wraper(callback):
        return ListeningComponent(callback, messages, users, component_type, check, custom_id)
    return wraper


class AlreadyDeferred(Exception):
    """This type of exception is thrown whenever the user tries to defer an interaction which was already deferred"""

    def __init__(self, *args: object) -> None:
        super().__init__("Interaction was already deferred")
class EphemeralDeletion(Exception):
    """This exception is thrown whenever the user tries to delete an ephemeral message (cannot be deleted)"""

    def __init__(self, *args: object) -> None:
        super().__init__("Cannot delete an ephemeral message")
class MissingOptionParameter(Exception):
    """This exception is thrown whenever a callback is missing a parameter which was specified in the slash command"""

    def __init__(self, option_name, *args: object) -> None:
        super().__init__("Missing parameter '" + option_name + "' in callback function")
class OptionalOptionParameter(Exception):
    """This exception is thrown whenever a callback function has a required parameter which is marked optional in the slash command"""

    def __init__(self, param_name, *args: object) -> None:
        super().__init__("Parameter '" + param_name + "' in callback function needs to be optional (" + param_name + "=None)")
class NoAsyncCallback(Exception):
    """This exception is thrown whenever a sync callback was provided"""
    
    def __init__(self, *args: object) -> None:
        super().__init__("callback has to be async")
class CallbackMissingContextCommandParameters(Exception):
    """This exception is thrown whenever a callback is missing the context parmeters"""
    
    def __init__(self, *args: object) -> None:
        super().__init__("Callback function for context commands has to accept 2 parameters (the used command, the message/user on which the interaction was used)")



from ..tools import MISSING, get, setup_logger
from ..http import BetterRoute, handle_rate_limit

from discord.ext import commands as com
from discord.errors import Forbidden, HTTPException, NotFound

import aiohttp

logging = setup_logger(__name__)

async def get_command(command_name, client: com.Bot, guild_id=None):
    return get(
        (
            await get_global_commands(client)
        ) if guild_id is None else (
            await get_guild_commands(client, guild_id)
        ),
        command_name, lambda x: x.get("name")
    )
async def get_id(command_name, client: com.bot, guild_id=None):
    found = (await get_command(command_name, client, guild_id))
    if found is None:
        raise Exception("No slash command found with name '" + command_name + "'")
    return found.get('id')

async def delete_global_commands(client: com.Bot):
    commands = await client.http.request(BetterRoute("GET", f"/applications/{client.user.id}/commands"))
    for x in commands:
        await delete_global_command(client, x["id"])
async def delete_guild_commands(client: com.Bot, guild_id):
    try:
        commands = await client.http.request(BetterRoute("GET", f"/applications/{client.user.id}/guilds/{guild_id}/commands"))
        for x in commands:
            await delete_guild_command(client, x["id"], guild_id)
    except Forbidden:
        logging.warn("got forbidden in " + str(guild_id))

async def delete_global_command(client: com.Bot, command_id):
    try:
        return await client.http.request(BetterRoute("DELETE", f"/applications/{client.user.id}/commands/{command_id}"))
    except HTTPException as ex:
        if ex.status == 429:
            await handle_rate_limit(await ex.response.json())
            return await delete_global_command(client, command_id)
        raise ex
async def delete_guild_command(client: com.Bot, command_id, guild_id):
    try:
        return await client.http.request(BetterRoute("DELETE", f"/applications/{client.user.id}/guilds/{guild_id}/commands/{command_id}"))
    except HTTPException as ex:
        if ex.status == 429:
            await handle_rate_limit(await ex.response.json())
            return await delete_guild_command(client, command_id, guild_id)
        else:
            raise ex
    except Exception as ex:
        raise ex

async def get_command_permissions(client: com.Bot, command_id, guild_id):
    try:
        return await client.http.request(BetterRoute("GET", f"/applications/{client.user.id}/guilds/{guild_id}/commands/{command_id}/permissions"))
    except NotFound:
        return {"id": command_id, "application_id": client.user.id, "permissions": []}
    except HTTPException as ex:
        if ex.status == 429:
            await handle_rate_limit(await ex.response.json())
            return await get_command_permissions(client, command_id, guild_id)
        else:
            raise ex
async def update_command_permissions(application_id, token, guild_id, command_id, permissions):
    async with aiohttp.ClientSession() as client:
        async with client.put(f"https://discord.com/api/v9/applications/{application_id}/guilds/{guild_id}/commands/{command_id}/permissions", 
            headers={"Authorization": "Bot " + token}, json={"permissions": permissions}) as response:
            if response.status == 200:
                return await response.json()
            elif response.status == 429:
                data = await handle_rate_limit(await response.json())
                await update_command_permissions(application_id, token, guild_id, command_id, permissions)
                return data
            raise HTTPException(response, response.content)

async def create_global_command(command: dict, client: com.Bot):
    try:
        return await client.http.request(BetterRoute("POST", f"/applications/{client.user.id}/commands"), json=command)
    except HTTPException as ex:
        if ex.status == 429:
            await handle_rate_limit(await ex.response.json())
            return await create_global_command(command, client)
        raise ex
async def create_guild_command(command, client: com.Bot, guild_id, permissions = []):
    try:
        data = await client.http.request(BetterRoute("POST", f"/applications/{client.user.id}/guilds/{guild_id}/commands"), json=command)
        await update_command_permissions(client.user.id, client.http.token, guild_id, data["id"], permissions)
        return data
    except HTTPException as ex:
        if ex.status == 429:
            await handle_rate_limit(await ex.response.json())
            return await create_guild_command(command, client, guild_id, permissions)
        raise ex


async def edit_global_command(command_id: str, client: com.Bot, new_command: dict):
    try:
        return await client.http.request(BetterRoute("PATCH", f"/applications/{client.user.id}/commands/{command_id}"), json=new_command)
    except HTTPException as ex:
        if ex.status == 429:
            await handle_rate_limit(await ex.response.json())
            return await edit_global_command(command_id, client, new_command)
        raise ex
async def edit_guild_command(command_id, client: com.Bot, guild_id: str, new_command: dict, permissions: dict=None):
    try:
        data = await client.http.request(BetterRoute("PATCH", f"/applications/{client.user.id}/guilds/{guild_id}/commands/{command_id}"), json=new_command)
        if permissions is not None:
            return await update_command_permissions(client.user.id, client.http.token, guild_id, data["id"], permissions)
    except HTTPException as ex:
        if ex.status == 429:
            await handle_rate_limit(await ex.response.json())
            return await edit_guild_command(command_id, client, guild_id, new_command, permissions)
        raise ex

async def get_global_commands(client):
    try:
        return await client.http.request(BetterRoute("GET", f"/applications/{client.user.id}/commands"))
    except HTTPException as ex:
        if ex.status == 429:
            await handle_rate_limit(await ex.response.json())
            return await get_global_commands(client)
        raise ex
async def get_guild_commands(client, guild_id):
    try:
        return await client.http.request(BetterRoute("GET", f"/applications/{client.user.id}/guilds/{guild_id}/commands"))
    except HTTPException as ex:
        if ex.status == 429:
            await handle_rate_limit(await ex.response.json())
            return await get_guild_commands(client, guild_id)
        if ex.status == 403:
            logging.warning("got forbidden in " + str(guild_id))
            return []
        raise ex



"""https://github.com/discord-py-ui/discord-ui/blob/main/discord_ui/override.py

    This module overrides some methods of the discord's functions.
    This overrides the Messageable.send, Webhook.send, the Message.__new__ method (whenever a new Message is created, it will use our own Message type)
    The same goes for the WebhookMessage, it will be overriden by our own Webhook type.
    And last but not least, if you're using dpy 2, the discord.ext.commands.Bot will be overriden with our
    own class, which enables `enable_debug_events` in order for our lib to work
"""

import asyncio
from .tools import MISSING
from .receive import Message, WebhookMessage
from .http import jsonifyMessage, BetterRoute, send_files

import discord
from discord.ext import commands

import sys

def override_dpy2_client():
    module = sys.modules["discord"]

    class OverridenV2Bot(commands.bot.Bot):
        """A overriden client that enables `enable_debug_events` for receiving the events"""
        def __init__(self, command_prefix, help_command = None, description = None, **options):
            commands.bot.Bot.__init__(self, command_prefix, help_command=help_command, description=description, enable_debug_events=True, **options)

    def client_override(cls, *args, **kwargs):
        if cls is commands.bot.Bot:
            return object.__new__(OverridenV2Bot)
        else:
            return object.__new__(cls)

    if discord.__version__.startswith("2"):
        module.ext.commands.bot.Bot.__new__ = client_override
    sys.modules["discord"] = module
def override_dpy():
    """This method overrides dpy methods. You shouldn't need to use this method by your own, the lib overrides everything by default"""
    module = sys.modules["discord"]

    #region message override
    async def send(self: discord.TextChannel, content=None, **kwargs) -> Message:
        channel = await self._get_channel()
        route = BetterRoute("POST", f"/channels/{channel.id}/messages")
        
        listener = None
        if kwargs.get("listener") is not None:
            listener = kwargs.pop("listener")
        if kwargs.get("components") is None and listener is not None:
            kwargs["components"] = listener.to_components()
        r = None
        if kwargs.get("file") is None and kwargs.get("files") is None:
            payload = jsonifyMessage(content=content, **kwargs)
            r = await self._state.http.request(route, json=payload)
        else:
            if kwargs.get("file") is not None:
                files = [kwargs.pop("file")]
            elif kwargs.get("files") is not None:
                files = kwargs.pop("files")
            
            payload = jsonifyMessage(content=content, **kwargs)
            r = await send_files(route, files=files, payload=payload, http=self._state.http)
        
        msg = Message(state=self._state, channel=channel, data=r)
        if kwargs.get("delete_after") is not None:
            await msg.delete(delay=kwargs.get("delete_after"))
    
        if listener is not None:
            listener._start(msg)

        return msg
    def message_override(cls, *args, **kwargs):
        if cls is discord.message.Message:
            return object.__new__(Message)
        else:
            return object.__new__(cls)


    module.abc.Messageable.send = send
    module.message.Message.__new__ = message_override
    #endregion

    #region webhook override
    def webhook_message_override(cls, *args, **kwargs):
        if cls is discord.webhook.WebhookMessage:
            return object.__new__(WebhookMessage)
        else:
            return object.__new__(cls)
    def send_webhook(self: discord.Webhook, content=MISSING, *, wait=False, username=MISSING, avatar_url=MISSING, tts=False, files=None, embed=MISSING, embeds=MISSING, allowed_mentions=MISSING, components=MISSING):
        payload = jsonifyMessage(content, tts=tts, embed=embed, embeds=embeds, allowed_mentions=allowed_mentions, components=components)

        if username is not None:
            payload["username"] = username
        if avatar_url is not None:
            payload["avatar_url"] = str(avatar_url)
        
        return self._adapter.execute_webhook(payload=payload, wait=wait, files=files)

    module.webhook.Webhook.send = send_webhook
    module.webhook.WebhookMessage.__new__ = webhook_message_override
    #endregion


    sys.modules["discord"] = module



"""
discord-ui extension
~~~~~~~~~~~~~~~~~~~~

A discord.py extension for discord's ui features like Buttons, SelectMenus, LinkButtons slash-commands and context-commands (message-commands and user-commands)

This libary features sending components, creating application-commands and receiving them

- - -

Links
    [Docs](https://discord-ui.rtfd.io/) | [Github](https://github.com/discord-py-ui/discord-ui/) | [PyPi](https://pypi.org/project/discord-ui/) | [License](https://github.com/git/git-scm.com/blob/main/MIT-LICENSE.txt)

- Made by [404kuso](https://github.com/404kuso) and [RedstoneZockt](https://github.com/RedstoneZockt)
- Made for [discord.py](https://github.com/Rapptz/discord.py) and you

- - -

### Issues, Bugs, etc.

If you find any issues, bugs, problems or anything, please report them to our [github](https://github.com/discord-py-ui/discord-ui/issues/) so we can fix them

### Ideas

If you have ideas for this package, plz feel free to tell us

### Help

If you need any help or assist, join our [discord](https://discord.gg/bDJCGD994p)

"""


from .client import Components, Slash, UI
from .components import ActionRow, Button, LinkButton, SelectMenu, SelectOption, ButtonStyles
from .slash.types import OptionType, SlashPermission, SlashOption
from .slash.tools import ParseMethod
from .tools import components_to_dict
from .slash.tools import create_choice
from .receive import Interaction, InteractionType, Message, WebhookMessage, PressedButton, SelectedMenu, ComponentContext, SlashedCommand, SlashedSubCommand, EphemeralMessage, EphemeralResponseMessage
from .slash import ext
from .listener import Listener

from .override import override_dpy, override_dpy2_client
override_dpy2_client()


__title__ = "discord-ui"
__version__ = "4.3.8"



import os
import setuptools


def get_version():
    """returns the version of the package"""
    with open("./discord_ui/__init__.py", "r", encoding="utf-8") as f:
        return [f for f in f.readlines() if f.startswith("__version__")][0].split('"')[1].split('"')[0]
def get_name():
    """returns the name of the package"""
    with open("./discord_ui/__init__.py", "r", encoding="utf-8") as f:
        return [f for f in f.readlines() if f.startswith("__title__")][0].split('"')[1].split('"')[0]
def get_readme():
    """returns the readme content for the package"""
    with open("./README.md", "r", encoding="utf-8") as f:
        return f.read()


setuptools.setup(
    name=get_name(),
    version=get_version(),
    project_urls={
        "Documentation": "https://discord-ui.rtfd.io/en/latest/",
        "Issue tracker": "https://github.com/discord-py-ui/discord-ui/issues",
    },
    author="404kuso, RedstoneZockt",
    author_email="bellou9022@gmail.com, redstoneprofihd@gmail.com",
    description="A disord.py extension for discord's ui/interaction features",
    long_description=get_readme(),
    long_description_content_type="text/markdown",
    url="https://github.com/discord-py-ui/discord-ui",
    packages=setuptools.find_packages(),
    python_requires='>=3.6',
    classifiers=[
        "Programming Language :: Python :: 3",
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Internet',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
        'Typing :: Typed'
    ]
)



"""
discord_ui.ext
~~~~~~~~~~~~~~~

An extension module to the libary that has some usefull decorators and functions 
for application-commands.

.. code-block::

    from discord_ui import ext

- - -

Important: Every decorator should be placed before the actual slashcommand decorator, 
except check auto-response decorators, they have to be placed after the target check decorator.

Note: **Theoretically**, these decorators could be used together with normal cog commands and 
not only the slash cog commands, but if you use them for the normal commands, the ``hidden`` param 
doesn't work
"""

import functools
import inspect

from discord.ext.commands import errors


def check_failure_response(content=None, hidden=False, **fields):
    """A decorator for autoresponding to a cog check that failed.
    
    The decorator has to be placed after the check deceorator

    Parameters
    ----------
    content: :class:`str`
        The raw message content
    tts: :class:`bool`
        Whether the message should be send with text-to-speech
    embed: :class:`discord.Embed`
        Embed rich content
    embeds: List[:class:`discord.Embed`]
        A list of embeds for the message
    file: :class:`discord.File`
        The file which will be attached to the message
    files: List[:class:`discord.File`]
        A list of files which will be attached to the message
    nonce: :class:`int`
        The nonce to use for sending this message
    allowed_mentions: :class:`discord.AllowedMentions`
        Controls the mentions being processed in this message
    mention_author: :class:`bool`
        Whether the author should be mentioned
    components: List[:class:`~Button` | :class:`~LinkButton` | :class:`~SelectMenu`]
        A list of message components to be included
    delete_after: :class:`float`
        After how many seconds the message should be deleted, only works for non-hiddend messages
    hidden: :class:`bool`
        Whether the response should be visible only to the user 
    

    Example
    -------

    .. code-block::

        # first check
        @ext.check_failure_response("command is guild only", hidden=True)
        @commands.guild_only()
        # second check
        @ext.check_failure_response("command can only used in nsfw channels", hidden=True)
        @commands.is_nsfw()
        @cogs.slash_cog(guild_ids=[867699578034716683])
        async def callback(self, ctx):
            ...
    """
    def wraper(cog):
        # get last check
        check = cog.checks[-1]
        _invoke = cog.invoke

        async def invoke(ctx, *args, **kwargs):
            try:
                if not check(ctx):
                    await ctx.send(content, **fields, hidden=hidden)
                    return
            except errors.CheckFailure:
                await ctx.send(content, **fields, hidden=hidden)
                raise
            await _invoke(ctx, *args, **kwargs)

        cog.invoke = invoke
        return cog
    return wraper

def any_failure_response(content, hidden=False, **fields):
    """Decorator for autoresponding to all checks of a cog command that failed.

    Note if you use the `check_failure_response` for a check, its response will be sent
    
    Parameters
    ----------
    content: :class:`str`
        The raw message content
    tts: :class:`bool`
        Whether the message should be send with text-to-speech
    embed: :class:`discord.Embed`
        Embed rich content
    embeds: List[:class:`discord.Embed`]
        A list of embeds for the message
    file: :class:`discord.File`
        The file which will be attached to the message
    files: List[:class:`discord.File`]
        A list of files which will be attached to the message
    nonce: :class:`int`
        The nonce to use for sending this message
    allowed_mentions: :class:`discord.AllowedMentions`
        Controls the mentions being processed in this message
    mention_author: :class:`bool`
        Whether the author should be mentioned
    components: List[:class:`~Button` | :class:`~LinkButton` | :class:`~SelectMenu`]
        A list of message components to be included
    delete_after: :class:`float`
        After how many seconds the message should be deleted, only works for non-hiddend messages
    hidden: :class:`bool`
        Whether the response should be visible only to the user 
    
    
    """
    def wraper(cog):
        _invoke = cog.invoke

        async def invoke(ctx, *args, **kwargs):
            try:
                if not await cog.can_run(ctx):
                    await ctx.send(content, hidden=hidden, **fields)
                    return
            except errors.CheckFailure:
                await ctx.send(content, hidden=hidden, **fields)
                raise
            await _invoke(ctx, *args, **kwargs)

        cog.invoke = invoke
        return cog
    return wraper

def guild_change(guild_id, *, name=None, description=None, default_permission=True):
    """A decorator for slashcommands that will apply changes to a specific guild

    Note that this decorator should mainly be used for guild commands, because if used with
    a global command, both commands will show up, the changed one and the global one.
    
    Parameters
    ----------
    guild_id: :class:`int` | :class:`str`
        The guild_id where the changes should be applied to
    name: :class:`str`, optional
        The new name; default None
    description: :class:`str`, optional
        The new description; default None
    default_permission: :class:`bool`, optional
        Whether this command can be used by default; default True
    
    """
    def wraper(callback):
        if not hasattr(callback, "__guild_changes__"):
            callback.__guild_changes__ = {}
        callback.__guild_changes__[str(guild_id)] = (name, description, default_permission)
        return callback
    return wraper

def alias(aliases):
    """Decorator for slashcommand aliases that will add the same command but with different names.
    
    Parameters
    ----------
    aliases: List[:class:`str`] | :class:`str`
        The alias(es) for the command with wich the command can be invoked with

    Usage:
    
    .. code-block::
    
        @ui.slash.command(name="cats", ...)
        @ui.slash.alias(["catz", "cute_things"])
    
    """
    def wraper(command):
        if not hasattr(command, "__aliases__"):
            command.__aliases__ = []
        # Allow multiple alias decorators
        command.__aliases__.extend(aliases if not isinstance(aliases, str) else [aliases])
        return command
    return wraper

def no_sync():
    """Decorator that will prevent an application-command to be synced with the api.
   
    Example
    -------

    .. code-block::

        from discord_ui import ext

        @ui.slash.command()
        @ext.no_sync()
        async def no_sync(ctx):
            \"\"\"This command will never be synced with the api\"\"\"
            ...

    """
    def wraper(callback):
        callback.__sync__ = False
        return callback
    return wraper

def auto_defer(hidden=False):
    """A decorator for auto deferring a interaction. This decorator has to be placed before the main decorator
    
    Parameters
    ----------
    hidden: :class:`bool`, optional
        Whether the interaction should be deferred hidden; default ``False``

    Example
    --------

    .. code-block::

        from discord_ui import ext

        @ui.slash.command()
        @ext.auto_defer()
        async def my_command(ctx):
            \"\"\"This command will be deferred automatically\"\"\"
            ...
    
    """
    # https://stackoverflow.com/questions/69076152/how-to-inject-a-line-of-code-into-an-existing-function#answers-header
    def decorator(func):
        func.__auto_defer__ = True
        @functools.wraps(func)
        async def wraper(*args, **kwargs):
            # if there is self param use the next one
            ctx = args[1 if list(inspect.signature(func).parameters.keys())[0] == "self" else 0]
            # use defer for "auto_defering"
            await ctx.defer(hidden=hidden)
            return await func(*args, **kwargs)
        return wraper
    return decorator